from __future__ import annotations

from enum import StrEnum
from pathlib import Path

from pydantic import BaseModel, model_validator


class SonarrQuality(StrEnum):
    UNKNOWN = "Unknown"
    SDTV = "SDTV"
    DVD = "DVD"
    WEBDL_480P = "WEBDL-480p"
    HDTV_720P = "HDTV-720p"
    WEBDL_720P = "WEBDL-720p"
    BLURAY_720P = "Bluray-720p"
    WEBDL_1080P = "WEBDL-1080p"
    BLURAY_1080P = "Bluray-1080p"
    HDTV_1080P = "HDTV-1080p"
    RAW_HD = "Raw-HD"
    HDTV_2160P = "HDTV-2160p"
    WEBDL_2160P = "WEBDL-2160p"
    BLURAY_2160P = "Bluray-2160p"
    BLURAY_1080P_REMUX = "Bluray-1080p Remux"
    BLURAY_2160P_REMUX = "Bluray-2160p Remux"
    WEBRIP_480P = "WEBRip-480p"
    WEBRIP_720P = "WEBRip-720p"
    WEBRIP_1080P = "WEBRip-1080p"
    WEBRIP_2160P = "WEBRip-2160p"


class SonarrConfig(BaseModel):
    base_url: str
    api_key: str
    shows_root: Path | None = None
    local_shows_root: Path | None = None

    @model_validator(mode="after")
    def check_path_map_consistency(self) -> SonarrConfig:
        if (self.shows_root is None) != (self.local_shows_root is None):
            raise ValueError(
                "shows_root and local_shows_root must both be set or both omitted"
            )
        return self


class SeasonConfig(BaseModel):
    season_offset: int = 0
    episode_offset: int = 0


class ShowConfig(BaseModel):
    name: str
    path: Path
    seasons: dict[int, SeasonConfig]
    sonarr_quality: SonarrQuality | None = None


class Config(BaseModel):
    shows: list[ShowConfig]
    sonarr: SonarrConfig | None = None


class FileOperation(BaseModel):
    source: Path
    destination: Path
    season: int
    episode: int


class UndoEntry(BaseModel):
    source: Path
    destination: Path



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any

import yaml

from .models import Config


APP_NAME: str = "seasonal-shift"


def get_config_dir() -> Path:
    xdg_config_home: str | None = os.getenv("XDG_CONFIG_HOME")

    if xdg_config_home:
        return Path(xdg_config_home) / APP_NAME

    return Path.home() / ".config" / APP_NAME


def find_default_config() -> Path:
    config_dir: Path = get_config_dir()

    yaml_path: Path = config_dir / "config.yaml"
    yml_path: Path = config_dir / "config.yml"
    json_path: Path = config_dir / "config.json"

    if yaml_path.exists():
        return yaml_path

    if yml_path.exists():
        return yml_path

    if json_path.exists():
        return json_path

    raise FileNotFoundError(
        f"No config file found in {config_dir}. "
        f"Expected config.yaml, config.yml, or config.json"
    )


def load_config(path: Path) -> Config:
    data: dict[str, Any]

    if path.suffix in (".yaml", ".yml"):
        data = yaml.safe_load(path.read_text())
    else:
        data = json.loads(path.read_text())

    return Config.model_validate(data)



from __future__ import annotations

import re
from collections.abc import Callable, Iterator
from pathlib import Path
from typing import Any

from rich import print
from watchdog.events import FileCreatedEvent, FileMovedEvent, FileSystemEventHandler
from watchdog.observers import Observer

from .executor import execute_operations, get_default_undo_file
from .models import Config, FileOperation, ShowConfig
from .planner import plan_operations
from .scanner import EPISODE_PATTERN


def _scan_single(file: Path) -> Iterator[tuple[Path, re.Match[str]]]:
    match = EPISODE_PATTERN.match(file.name)
    if match:
        yield file, match


class EpisodeEventHandler(FileSystemEventHandler):
    def __init__(
        self,
        show: ShowConfig,
        sonarr_cb: Callable[[ShowConfig, list[FileOperation]], Any] | None,
    ) -> None:
        self._show = show
        self._sonarr_cb = sonarr_cb

    def on_created(self, event: FileCreatedEvent) -> None:  # type: ignore[override]
        if not event.is_directory:
            self._handle(Path(event.src_path))

    def on_moved(self, event: FileMovedEvent) -> None:  # type: ignore[override]
        if not event.is_directory:
            self._handle(Path(event.dest_path))

    def _handle(self, path: Path) -> None:
        operations = plan_operations(self._show, lambda _: _scan_single(path))
        if not operations:
            return
        undo_file = get_default_undo_file()
        execute_operations(operations, undo_file)
        for op in operations:
            print(f"[green]Watch:[/] {op.source.name} → {op.destination.name}")
        if self._sonarr_cb:
            self._sonarr_cb(self._show, operations)


def run_watch(
    cfg: Config,
    sonarr_cb: Callable[[ShowConfig, list[FileOperation]], Any] | None = None,
) -> None:
    observer = Observer()
    for show in cfg.shows:
        handler = EpisodeEventHandler(show, sonarr_cb)
        observer.schedule(handler, str(show.path), recursive=True)

    observer.start()
    print(f"[blue]Watching {len(cfg.shows)} show(s). Press Ctrl+C to stop.[/]")
    try:
        observer.join()
    except KeyboardInterrupt:
        observer.stop()
        observer.join()



from __future__ import annotations

import json
import time
from pathlib import Path
from typing import Any, NotRequired, Required, TypedDict

import requests


class Series(TypedDict):
    id: int
    title: str
    path: str


class Episode(TypedDict):
    id: int
    episodeNumber: int


class Language(TypedDict):
    id: int
    name: str


class QualityWrapper(TypedDict):
    quality: dict[str, Any]


class ManualImportCandidate(TypedDict):
    # fields Sonarr always returns
    path: Required[str]
    quality: Required[QualityWrapper]

    # fields we set manually
    seriesId: NotRequired[int]
    seasonNumber: NotRequired[int]
    episodeIds: NotRequired[list[int]]
    languages: NotRequired[list[Language]]


class SonarrClient:
    def __init__(self, base_url: str, api_key: str) -> None:
        self.base_url: str = base_url.rstrip("/")
        self.session: requests.Session = requests.Session()
        self.session.headers.update({"X-Api-Key": api_key})

    def _get(  # pyright: ignore[reportAny]
        self,
        endpoint: str,
        **kwargs: Any,  # pyright: ignore[reportExplicitAny, reportAny]
    ) -> Any:  # pyright: ignore[reportExplicitAny]
        r = self.session.get(
            f"{self.base_url}{endpoint}", **kwargs  # pyright: ignore[reportAny]
        )
        if not r.ok:
            # print response content for debugging
            print(f"Error response from Sonarr: {r.status_code} {r.text}")
            print(f"Reponse: {json.dumps(r.json(), indent=2)}")
        r.raise_for_status()
        return r.json()  # pyright: ignore[reportAny]

    def _post(  # pyright: ignore[reportAny]
        self,
        endpoint: str,
        **kwargs: Any,  # pyright: ignore[reportExplicitAny, reportAny]
    ) -> Any:  # pyright: ignore[reportExplicitAny]
        r = self.session.post(
            f"{self.base_url}{endpoint}", **kwargs  # pyright: ignore[reportAny]
        )
        r.raise_for_status()
        return r.json()  # pyright: ignore[reportAny]

    # ------------------------------------------------------------
    # Series
    # ------------------------------------------------------------

    def get_series(self, show_name: str) -> tuple[int, str]:
        series_list: list[Series] = self._get(  # pyright: ignore[reportAny]
            "/api/v3/series"
        )

        for series in series_list:
            if series["title"].lower() == show_name.lower():
                return series["id"], series["path"]

        raise RuntimeError(f"Series not found: {show_name}")

    # ------------------------------------------------------------
    # Episodes
    # ------------------------------------------------------------

    def get_quality_id(self, quality_name: str) -> int:
        definitions: list[dict[str, Any]] = self._get("/api/v3/qualitydefinition")
        for defn in definitions:
            if defn["quality"]["name"] == quality_name:
                return int(defn["quality"]["id"])
        raise RuntimeError(f"Quality not found in Sonarr: {quality_name!r}")

    def get_episode_id(self, series_id: int, season: int, episode: int) -> int:
        episodes: list[Episode] = self._get(  # pyright: ignore[reportAny]
            "/api/v3/episode",
            params={
                "seriesId": series_id,
                "seasonNumber": season,
            },
        )

        for ep in episodes:
            if ep["episodeNumber"] == episode:
                return ep["id"]

        raise RuntimeError(f"Episode not found S{season:02}E{episode:02}")

    # ------------------------------------------------------------
    # Manual Import
    # ------------------------------------------------------------

    def detect_file(
        self,
        series_id: int,
        file_path: str | Path,
        season: int,
        episode_id: int,
        *,
        quality_id: int = 17,
    ) -> ManualImportCandidate:

        path = Path(file_path)
        folder = str(path.parent)

        candidates: list[ManualImportCandidate] = (  # pyright: ignore[reportAny]
            self._get(
                "/api/v3/manualimport",
                params={
                    "folder": folder,
                    "seriesId": series_id,
                },
            )
        )

        if not candidates:
            raise RuntimeError("Sonarr detected no files")

        candidate = next(
            (c for c in candidates if c["path"] == str(path)),
            None,
        )

        if candidate is None:
            raise RuntimeError(f"File not detected by Sonarr: {path}")

        candidate["seriesId"] = series_id
        candidate["seasonNumber"] = season
        candidate["episodeIds"] = [episode_id]

        candidate["quality"]["quality"]["id"] = quality_id

        # candidate["languages"] = [
        #     {
        #         "id": 8,
        #         "name": "Japanese",
        #     }
        # ]

        result: list[ManualImportCandidate] = self._post(  # pyright: ignore[reportAny]
            "/api/v3/manualimport",
            json=[candidate],
        )

        return result[0]

    def import_file(
        self,
        series_id: int,
        episode_id: int,
        candidate: ManualImportCandidate,
    ) -> dict[str, Any]:

        candidate["seriesId"] = series_id
        candidate["episodeIds"] = [episode_id]

        payload: dict[str, Any] = {
            "name": "ManualImport",
            "files": [candidate],
            "importMode": "auto",
        }

        return self._post("/api/v3/command", json=payload)

    def wait_for_command(self, command_id: int, *, poll_interval: float = 2.0) -> None:
        while True:
            result: dict[str, Any] = self._get(f"/api/v3/command/{command_id}")
            status: str = result.get("status", "")
            if status == "completed":
                return
            if status == "failed":
                raise RuntimeError(
                    f"Sonarr command {command_id} failed: {result.get('message', '')}"
                )
            time.sleep(poll_interval)

    def refresh_series(self, series_id: int) -> dict[str, Any]:
        payload = {
            "name": "RefreshSeries",
            "seriesId": series_id,
        }

        return self._post("/api/v3/command", json=payload)



from __future__ import annotations

import re
from pathlib import Path
from collections.abc import Iterator

EPISODE_PATTERN = re.compile(
    r"(.+) - [Ss](\d{2})[Ee](\d{2})( - .*)?\.(.+)"
)


def scan_show(show_path: Path) -> Iterator[tuple[Path, re.Match[str]]]:
    for season_dir in sorted(show_path.iterdir()):

        if not season_dir.is_dir():
            continue

        for file in season_dir.iterdir():

            if not file.is_file():
                continue

            match = EPISODE_PATTERN.match(file.name)

            if match:
                yield file, match
